/*****************************************************************************************************************************
--
--    File Name    : cdcfifo.v

--    Description  : This module implements CDC fifo                                 

-- Targeted device : Microsemi-SoC                     
-- Author          : India Solutions Team

-- SVN Revision Information:
-- SVN $Revision: TBD
-- SVN $Date: TBD
--
--
--
-- COPYRIGHT 2016 BY MICROSEMI 
-- THE INFORMATION CONTAINED IN THIS DOCUMENT IS SUBJECT TO LICENSING RESTRICTIONS 
-- FROM MICROSEMI CORP.  IF YOU ARE NOT IN POSSESSION OF WRITTEN AUTHORIZATION FROM 
-- MICROSEMI FOR USE OF THIS FILE, THEN THE FILE SHOULD BE IMMEDIATELY DESTROYED AND 
-- NO BACK-UP OF THE FILE SHOULD BE MADE. 
-- 

--******************************************************************************************************************************/
`timescale 1 ns / 1 ps

module fifocdc
 #(
//---------------------------Parameters----------------------------------------
parameter [3:0]               ADDR_WIDTH = 12,
parameter [7:0]               DATA_WIDTH = 8
)
(
   k_lim,
   wrrstn,
   wrsrst,
   wrclk,
   wrreq,
   wrdata,
   wrfull,
   wrusedw,
   rdrstn,
   rdclk,
   rdsrst,
   rdreq,
   rddata,
   rdempty,
   rdusedw
);
 //  parameter [3:0]               ADDR_WIDTH = 12;
 //  parameter [7:0]               DATA_WIDTH = 32;
   input [ADDR_WIDTH-1:0]        k_lim;
   input                         wrrstn;
   input                         wrsrst;
   input                         wrclk;
   input                         wrreq;
   input [(DATA_WIDTH-1):0]      wrdata;
   output reg                    wrfull;
   output [(ADDR_WIDTH-1):0]     wrusedw;
   
   input                         rdrstn;
   input                         rdclk;
   input                         rdsrst;
   input                         rdreq;
   output [(DATA_WIDTH-1):0]     rddata;
   output                        rdempty;
   output reg [(ADDR_WIDTH-1):0] rdusedw;
   
   
   //-----------------------------------------------------------------------------
   parameter [31:0]              GRAY_P1 = 32'h00000001;		// +1 in gray code
   parameter [31:0]              GRAY_P2 = 32'h00000003;		// +2 in gray code
   
   wire                          rdallow_n;
   wire                          wrallow_n;
   wire                          rdempty_n;
   reg [(ADDR_WIDTH-1):0]        rdaddr_r;
   reg [(ADDR_WIDTH-1):0]        rdaddr_n /* synthesis syn_maxfan=10 */;
   reg [(ADDR_WIDTH-1):0]        rdaddr_sync_r;
   reg [(ADDR_WIDTH-1):0]        rdaddr_sync_rr;
   reg [(ADDR_WIDTH-1):0]        wraddr_r;
   reg [(ADDR_WIDTH-1):0]        wraddr_rr;
   reg [(ADDR_WIDTH-1):0]        wraddr_plus2_r;
   reg [(ADDR_WIDTH-1):0]        wraddr_plus1_r;
   reg [(ADDR_WIDTH-1):0]        wraddr_plus2_n;
   reg [(ADDR_WIDTH-1):0]        wraddr_sync_r;
   reg [(ADDR_WIDTH-1):0]        wraddr_sync_rr;
   reg [(ADDR_WIDTH-1):0]        wrusedw_r;
   wire [(ADDR_WIDTH-2):0]       zero;
   
   //--------------------------------------------------------------
   //  Dual-port RAM instantiation
   //--------------------------------------------------------------
   assign zero = {ADDR_WIDTH-1{1'b0}};
   
   
   fifocdcram #(.ADDR_WIDTH(ADDR_WIDTH), .DATA_WIDTH(DATA_WIDTH)) dcram(
      .data(wrdata),
      .wren(wrallow_n),
      .wraddress(wraddr_r),
      //.rdaddress(rdaddr_n),
      .rdaddress(rdaddr_r),
      .wrclock(wrclk),
      .rdclock(rdclk),
      .q(rddata)
   );
   
   //--------------------------------------------------------------
   //  Write control
   //--------------------------------------------------------------
   
   // Compute wraddr_plus2_n <=wraddr_plus2_r+'1'     
   
   always @(wraddr_plus2_r or zero)
   begin: WRADDR_PLUS_PROC
      reg [(ADDR_WIDTH-1):0]        tog;
      reg                           done;
      integer                       i;
      integer                       j;
      tog[(ADDR_WIDTH - 1)] = wraddr_plus2_r[(ADDR_WIDTH - 1)];
      tog[(ADDR_WIDTH - 2):0] = zero;
      done = 1'b0;
      
      for (i = 0; i <= (ADDR_WIDTH - 1); i = i + 1)
      begin
         for (j = i; j <= (ADDR_WIDTH - 1); j = j + 1)
            tog[i] = tog[i] ^ wraddr_plus2_r[j];
         wraddr_plus2_n[i] <= wraddr_plus2_r[i] ^ ((~done) & (~tog[i]));
         done = done | (~tog[i]);
      end
   end
   
   // Write process
   
   always @(negedge wrrstn or posedge wrclk)
   begin: WRITE_PROC
      reg [(ADDR_WIDTH-1):0]        wraddr_rr_bin;
      reg [(ADDR_WIDTH-1):0]        rdaddr_sync_rr_bin;
      integer                       i;
      if (wrrstn == 1'b0)
      begin
         wraddr_r <= {ADDR_WIDTH{1'b0}};
         wraddr_rr <= {ADDR_WIDTH{1'b0}};
         wraddr_plus1_r <= GRAY_P1[(ADDR_WIDTH - 1):0];
         wraddr_plus2_r <= GRAY_P2[(ADDR_WIDTH - 1):0];
         rdaddr_sync_r <= {ADDR_WIDTH{1'b0}};
         rdaddr_sync_rr <= {ADDR_WIDTH{1'b0}};
         wrusedw_r <= {ADDR_WIDTH{1'b0}};
         wrfull <= 1'b0;
         rdaddr_sync_rr_bin = {ADDR_WIDTH{1'b0}};
         wraddr_rr_bin = {ADDR_WIDTH{1'b0}};
      end
      else 
      begin
         if (wrsrst == 1'b1)
         begin
            wraddr_r <= {ADDR_WIDTH{1'b0}};
            wraddr_rr <= {ADDR_WIDTH{1'b0}};
            wraddr_plus1_r <= GRAY_P1[(ADDR_WIDTH - 1):0];
            wraddr_plus2_r <= GRAY_P2[(ADDR_WIDTH - 1):0];
            rdaddr_sync_r <= {ADDR_WIDTH{1'b0}};
            rdaddr_sync_rr <= {ADDR_WIDTH{1'b0}};
            wrusedw_r <= {ADDR_WIDTH{1'b0}};
            wrfull <= 1'b0;
            rdaddr_sync_rr_bin = {ADDR_WIDTH{1'b0}};
            wraddr_rr_bin = {ADDR_WIDTH{1'b0}};
         end
         else
         begin
            // Maintain write pointers
            if (wrallow_n == 1'b1)
            begin
               wraddr_plus2_r <= wraddr_plus2_n;
               wraddr_plus1_r <= wraddr_plus2_r;
               wraddr_r <= wraddr_plus1_r;
            end
            // Wait for one cycle in order to allow data to be written into RAM
            // (wraddr_r can be used instead if RAM latency is not an issue)
            wraddr_rr <= wraddr_r;
            // Resynchronize read address pointer
            rdaddr_sync_r <= rdaddr_r;
            rdaddr_sync_rr <= rdaddr_sync_r;
            // Gray to binary conversion
            wraddr_rr_bin[(ADDR_WIDTH - 1)] = wraddr_rr[(ADDR_WIDTH - 1)];
            for (i = (ADDR_WIDTH - 2); i >= 0; i = i - 1)
               wraddr_rr_bin[i] = wraddr_rr_bin[i + 1] ^ wraddr_rr[i];
            rdaddr_sync_rr_bin[(ADDR_WIDTH - 1)] = rdaddr_sync_rr[(ADDR_WIDTH - 1)];
            for (i = (ADDR_WIDTH - 2); i >= 0; i = i - 1)
               rdaddr_sync_rr_bin[i] = rdaddr_sync_rr_bin[i + 1] ^ rdaddr_sync_rr[i];
            if (wrusedw_r >= k_lim)
               wrfull <= 1'b1;
            else
               wrfull <= 1'b0;
            // Generate word count
            wrusedw_r <= wraddr_rr_bin - rdaddr_sync_rr_bin;
         end
      end
   end
   
   assign wrusedw = wrusedw_r;
   assign wrallow_n = wrreq;
   
   //--------------------------------------------------------------
   //  Read control
   //--------------------------------------------------------------
   
   // Compute rdaddr_n <=rdaddr_r+rdallow_n 
   
   always @(rdaddr_r or rdallow_n or zero)
   begin: READ_ADDR_PROC
      reg [(ADDR_WIDTH-1):0]        tog;
      reg                           done;
      integer                       i;
      integer                       j;
      tog[(ADDR_WIDTH - 1)] = rdaddr_r[(ADDR_WIDTH - 1)];
      tog[(ADDR_WIDTH - 2):0] = zero;
      done = 1'b0;
      for (i = 0; i <= (ADDR_WIDTH - 1); i = i + 1)
      begin
         for (j = i; j <= (ADDR_WIDTH - 1); j = j + 1)
            tog[i] = tog[i] ^ rdaddr_r[j];
         rdaddr_n[i] <= rdaddr_r[i] ^ (rdallow_n & (~done) & (~tog[i]));
         done = done | (~tog[i]);
      end
   end
   
   // Read process
   
   always @(negedge rdrstn or posedge rdclk)
   begin: READ_PROC
      reg [(ADDR_WIDTH-1):0]        rdaddr_r_bin;
      reg [(ADDR_WIDTH-1):0]        wraddr_sync_rr_bin;
      integer                       i;
      if (rdrstn == 1'b0)
      begin
         rdaddr_r <= {ADDR_WIDTH{1'b0}};
         wraddr_sync_r <= {ADDR_WIDTH{1'b0}};
         wraddr_sync_rr <= {ADDR_WIDTH{1'b0}};
         rdusedw <= {ADDR_WIDTH{1'b0}};
         rdaddr_r_bin = {ADDR_WIDTH{1'b0}};
         wraddr_sync_rr_bin = {ADDR_WIDTH{1'b0}};
      end
      else 
      begin
         if (rdsrst == 1'b1)
         begin
            rdaddr_r <= {ADDR_WIDTH{1'b0}};
            wraddr_sync_r <= {ADDR_WIDTH{1'b0}};
            wraddr_sync_rr <= {ADDR_WIDTH{1'b0}};
            rdusedw <= {ADDR_WIDTH{1'b0}};
            rdaddr_r_bin = {ADDR_WIDTH{1'b0}};
            wraddr_sync_rr_bin = {ADDR_WIDTH{1'b0}};
         end
         else
         begin
            // Maintain read pointers
            rdaddr_r <= rdaddr_n;
            // Resynchronise write address pointer
            wraddr_sync_r <= wraddr_rr;
            wraddr_sync_rr <= wraddr_sync_r;
            // Gray to binary conversion
            wraddr_sync_rr_bin[(ADDR_WIDTH - 1)] = wraddr_sync_rr[(ADDR_WIDTH - 1)];
            for (i = (ADDR_WIDTH - 2); i >= 0; i = i - 1)
               wraddr_sync_rr_bin[i] = wraddr_sync_rr_bin[i + 1] ^ wraddr_sync_rr[i];
            rdaddr_r_bin[(ADDR_WIDTH - 1)] = rdaddr_r[(ADDR_WIDTH - 1)];
            for (i = (ADDR_WIDTH - 2); i >= 0; i = i - 1)
               rdaddr_r_bin[i] = rdaddr_r_bin[i + 1] ^ rdaddr_r[i];
            // Compute read word count
            rdusedw <= wraddr_sync_rr_bin - rdaddr_r_bin;
         end
      end
   end
   
   assign rdempty_n = (wraddr_sync_rr == rdaddr_r) ? 1'b1 : 
                      1'b0;
   assign rdempty = rdempty_n;
   assign rdallow_n = rdreq & (~rdempty_n);
   
endmodule
